<?php

/**
 * @file
 * Code altering the custom UIs for many CTools exportable objects / entities.
 *
 * - various types supporting the CTools Export UI class: see
 *   _ctex_diff_ctools_plugin_post_alter()
 * - Entityform types, Mailchimp signups, Profile2 types & probably more,
 *   through hook_form_entity_ui_overview_form_alter()
 * - Boxes, FacetAPI, Feeds,  Path Breadcrumbs, SearchAPI entities, and
 *   Strongarm/variables through their own hook_form_alter implementations
 * - Page Manager pages & handlers, through their custom alter implementations
 * - Backup & Migrate crud types, Data tables, Display Suite view modes/fields,
 *   Metatags and OAuth contexts through hook_preprocess_table()
 * - Rules, Rules Components, Commerce Payment conf, through hook_page_alter()
 * - ... (more to come?)
 *
 * There are also modules that have no custom UI support (for exporting and
 * reverting) for their objects, so this could only be done through the
 * admin/structure/ctex_diff links we've added:
 * - ds_layout_settings (reverting makes no sense because it has a danger of
 *   going out of sync with other field display settings, maybe?)
 * - elysia_cron
 * - field_group: reverting makes no sense because each time a content type's
 *   field settings are saved, the field group will show as 'overridden' again.
 *
 * Special mentions:
 * - fe_block_settings is not in our global overview but is visible in features.
 *   fe_block_boxes is in our global overview but is not visible in features.
 *   Until someone is interested in features_extra module, this will not be
 *   checked.
 *
 * Non-core modules with Features integration but not CTools, so we cannot
 * support them (and don't need to; if you have Features, you have a UI):
 * - conditional_fields
 * - flag (has its own custom export code/UI too.)
 * - paragraphs
 */

/**
 * Implements hook_ctools_plugin_post_alter().
 *
 * Adds 'diff' to the operations of a CTools export ui plugin.
 */
function _ctex_diff_ctools_plugin_post_alter(&$plugin, &$info) {

  // This defines the ctools objects which have a CTools UI based admin form,
  // where we want to insert a 'diff' link: Schemas, and the resulting paths.
  // (The path value is not actually used, except for the position of '%'
  // inside it and the prefix before it. For the rest, it just acts as
  // documentation: it's the path which the code below will make work - plus the
  // path where 'diff' is 'code'.)
  //
  // NOTE: after changing these, the menu cache must be cleared.
  $supported = array(
    'context' => 'admin/structure/context/list/%/diff',
    'dfp_tags' => 'admin/structure/dfp_ads/list/%/diff',
    'ctools_custom_content' => 'admin/structure/ctools-content/list/%/diff',
    'entityqueue_queue' => 'admin/structure/entityqueue/list/%/diff',
    'flexslider_optionset' => 'admin/config/media/flexslider/list/%/diff',
    'page_manager_existing_pages' => 'admin/structure/pages/existing-pages/list/%/diff',
    'panels_layout' => 'admin/structure/panels/layouts/list/%/diff',
    'panels_mini' => 'admin/structure/mini-panels/list/%/diff',
    'services_endpoint' => 'admin/structure/services/list/%/diff',
    'slick_optionset' => 'admin/config/media/slick/list/%/diff',
    'vef_video_styles' => 'admin/config/media/vef_video_styles/list/%/diff',
    'views_view' => 'admin/structure/views/view/%/diff',
  );

  if (isset($plugin['plugin type']) && $plugin['plugin type'] === 'export_ui'
      && isset($plugin['schema']) && isset($supported[$plugin['schema']])
      && isset($plugin['allowed operations'])) {

    // 1)
    // Add the extra operation to the ctools plugin; this is something that
    // should be done on every page load. (Plugin definition isn't cached.)
    $plugin['allowed operations']['diff'] = array('title' => t('Diff'));
    $plugin['allowed operations']['code'] = array('title' => t('View default'));

    // 2)
    // Register menu path so that the link will actually work. This only has
    // effect when the menu cache is being rebuilt.

    // Find the location of '%' in the menu path.
    $path_components = explode('/', $supported[$plugin['schema']]);
    $position = array_search('%', $path_components);
    // Use a variable prefix so e.g. Views will use 'view' instead of 'list'
    // just like its own other links, for consistency. It isn't strictly
    // necessary but hey why not.
    $view_prefix = $path_components[$position - 1];

    $plugin['menu']['items'] += array('diff' => array(
      // The path is relative, so we can't refer to 'our own' menu path. We use
      // the below. We must use %ctools_export_ui in 'path', not % (i.e. there
      // is a mandatory 'load function'); see also ctex_diff_ctools_ui_title().
      'path' => "$view_prefix/%ctools_export_ui/diff",
      // The load function (ctools_export_ui_load()) needs $plugin_name as a 2nd
      // argument; provide that here.
      'load arguments' => array($plugin['name']),
      'type' => MENU_CALLBACK,
      'page callback' => 'ctex_diff_ctools_ui_object_diff',
      'page arguments' => array($plugin['schema'], $position),
      'title callback' => 'ctex_diff_ctools_ui_title',
      'title arguments' => array($plugin['schema'], $position, 'diff'),
      'access callback' => 'user_access',
      'access arguments' => array('view ctools export definitions'),
      // Note that unlike e.g. core menus, the value of the permission does not
      // have any effect on the link being output. The access callback is only
      // checked when visiting the menu. To suppress outputting the 'diff' link
      // if an item is not overridden, see the submit callback below.
    ));

    $plugin['menu']['items'] += array('code' => array(
      'path' => "$view_prefix/%ctools_export_ui/code",
      'load arguments' => array($plugin['name']),
      'type' => MENU_CALLBACK,
      'page callback' => 'ctex_diff_ctools_ui_show_code',
      'page arguments' => array($plugin['schema'], $position),
      'title callback' => 'ctex_diff_ctools_ui_title',
      'title arguments' => array($plugin['schema'], $position, 'default definition'),
      'access callback' => 'user_access',
      'access arguments' => array('view ctools export definitions'),
    ));
  }
}

/**
 * 'Quasi submit' callback for the CTools Export UI list form.
 *
 * The form is always processed and this 'submit' handler is always called, also
 * when just building the list for the first time without anything being posted.
 */
function _ctex_diff_ctools_export_ui_list_form_submit($form, $form_state) {

  // If the below does not do anything, then probably module weights should be
  // changed so this is the last submit function run.

  // Situation description:
  // ctools_export_ui::list_form_submit() -> build_operations() has just
  // collected all 'allowed operations' directly from the plugin class, and
  // list_build_row() immediately rendered it into HTML. The logic of
  // dynamically disabling the default operations is hardcoded there, and there
  // is no way of doing the same for our extra operations.
  // (...or at least, there's no way that is in line with the above alter hook.
  // We could subclass ctools_export_ui but that isn't really an option because
  // we'd have to subclass/extend all existing subclasses like views_ui, too.
  // And re-register all plugins.)
  // So we end up with rendered HTML links; the only way to 'disable' links
  // per-item is to change that HTML.

  // The below has many doublechecks if all data structure are as expected. If
  // not, we do nothing; keeping the links is never dangerous.
  if (isset($form_state['object']->rows) && is_array($form_state['object']->rows)
      && isset($form_state['object']->items) && is_array($form_state['object']->items)) {

    foreach ($form_state['object']->items as $name => $item) {
      if (isset($item->export_type)
          && $item->export_type != (EXPORT_IN_CODE | EXPORT_IN_DATABASE)) {
        // This item is not overridden, so we want to remove 'diff/code' links.
        if (($row_key = _ctex_diff_ctools_export_ui_find_row($name, $form_state['object']->rows))
            !== NULL) {
          $row = $form_state['object']->rows[$row_key];

          // Abovementioned methods put the rendered 'operations' in the last
          // column.
          $last_column = end($row['data']);
          $key = key($row['data']);
          if (isset($last_column['class']) && is_array($last_column['class'])
              // seen: ctools-export-ui-operations, views-operations.
              && strpos(reset($last_column['class']), '-operations')
              && isset($last_column['data'])) {
            // Find and remove 'diff' and 'code' links from rendered HTML.
            $last_class_removed = FALSE;
            $m = array();
            foreach (array('diff', 'code') as $code) {
              if (preg_match('|<li class="' . $code . '(.*?)"><a href=.*?</a></li>|', $last_column['data'], $m)) {
                if (in_array('last', explode(' ', $m[1]), TRUE)) {
                  $last_class_removed = TRUE;
                }
                $form_state['object']->rows[$row_key]['data'][$key]['data'] =
                  preg_replace('|<li class="' . $code . '.*?"><a href=.*?</a></li>|',
                    '',
                    $form_state['object']->rows[$row_key]['data'][$key]['data']);
              }
            }
            if ($last_class_removed) {
              // Place 'last' class back inside the last <li>.
              $string = $form_state['object']->rows[$row_key]['data'][$key]['data'];
              if (($pos = strrpos($string, '<li class="')) !== FALSE
                  && $pos2 = strpos($string, '"', $pos + 11)) {
                $substr = substr($string, $pos, $pos2 - $pos);
                $form_state['object']->rows[$row_key]['data'][$key]['data'] = str_replace($substr, $substr . ' last', $string);
              }
            }
          }
        }
      }
    }
  }
}

/**
 * Finds a HTML row matching a machine name which also has a 'data' value.
 *
 * @param string $name
 *   The machine to search for.
 * @param array $rows
 *   The (CTools export ui) rendered HTML rows to search in.
 *
 * @return string|null
 *   The key of the row; NULL if not found.
 */
function _ctex_diff_ctools_export_ui_find_row($name, array $rows) {
  // All we know so far is it isn't exact. For instance, context can have
  // "< Untagged >:$name" as row keys, and "<Untagged >" can probably be
  // replaced by semi random strings.
  $match = NULL;
  if (isset($rows[$name])) {
    $match = $name;
  }
  else {
    foreach ($rows as $key => $row) {
      if (in_array($name, explode(':', $key), TRUE)) {
        $match = $key;
        break;
      }
    }
  }

  return isset($match) && isset($rows[$match]['data']) && is_array($rows[$match]['data'])
    ? $match : NULL;
}

/**
 * Implements hook_page_manager_operations_alter().
 */
function _ctex_diff_page_manager_operations_alter(&$result, $page) {
  if (!isset($result['actions']['children']['diff'])
      && isset($page->subtask_id)) {
    $item = ctools_export_crud_load('page_manager_pages', $page->subtask_id);
    if (isset($item->export_type) && $item->export_type == (EXPORT_IN_CODE | EXPORT_IN_DATABASE)) {
      $result['actions']['children']['diff'] = array(
        'title' => t('Diff'),
        'description' => t('Differences between default and overridden version.'),
        'form' => 'ctex_diff_page_manager_diff_form',
        // Get rid of update and save button which is bad here.
        'no update and save' => TRUE,
      );
      $result['actions']['children']['code'] = array(
        'title' => t('View default'),
        'description' => t('Default version.'),
        'form' => 'ctex_diff_page_manager_code_form',
        'no update and save' => TRUE,
      );
    }
  }
}

/**
 * Implements hook_page_manager_operations_alter().
 */
function _ctex_diff_page_manager_variant_operations_alter(&$result, $handler) {
  if (!isset($result['children']['actions']['children']['diff'])) {
    $item = ctools_export_crud_load('page_manager_handlers', $handler->name);
    if (isset($item->export_type) && $item->export_type == (EXPORT_IN_CODE | EXPORT_IN_DATABASE)) {
      $result['children']['actions']['children']['diff'] = array(
        'title' => t('Diff'),
        'description' => t('Differences between default and overridden version.'),
        'form' => 'ctex_diff_page_manager_diff_form',
        'no update and save' => TRUE,
      );
      $result['children']['actions']['children']['code'] = array(
        'title' => t('View default'),
        'description' => t('Default version.'),
        'form' => 'ctex_diff_page_manager_code_form',
        'no update and save' => TRUE,
      );
    }
  }
}

/**
 * Implements hook_form_FORM_ID_alter().
 */
function _ctex_diff_form_block_admin_display_form_alter(&$form, $form_state) {
  // Add links for Boxes. Logic copied from
  // boxes_form_block_admin_display_form_alter().
  foreach (element_children($form['blocks']) as $i) {
    if (isset($form['blocks'][$i]['module']['#value']) && $form['blocks'][$i]['module']['#value'] == 'boxes') {
      $delta = $form['blocks'][$i]['delta']['#value'];
      if (strpos($delta, 'boxes_add__') !== 0) {
        $box = boxes_box_load($delta);
        if (($box->export_type & EXPORT_IN_DATABASE) && ($box->export_type & EXPORT_IN_CODE)) {
          // The standard block-admin-display-form.tpl.php and its preprocess
          // function use/render very specufic properties so we can't just add
          // another property. Instead of that, we'll stick the extra link into
          // the 'delete' property, which the 'boxes' alter has filled just
          // before us with a link.
          $form['blocks'][$i]['delete'] = array('#markup' => render($form['blocks'][$i]['delete']) . ' | '
            . l(t('diff'), "admin/structure/ctex_diff/diff/box/$delta") . ' | '
            . l(t('view default'), "admin/structure/ctex_diff/code/box/$delta"));
        }
      }
    }
  }
}

/**
 * Implements hook_form_FORM_ID_alter().
 *
 * Adds 'more admin functions' underneath each table in bulk export form.
 */
function _ctex_diff_form_bulk_export_export_form_alter(&$form, &$form_state) {

  foreach ($form as $module => $array) {
    if (isset($array['tables']) && is_array($array['tables'])) {
      foreach ($array['tables'] as $tablename => $tabledef) {
        if (isset($tabledef['#suffix']) && $tabledef['#suffix'] == '</div>') {
          $form[$module]['tables'][$tablename]['#header'][0] .= ' <em>' . l(t('Diff options'), 'admin/structure/ctex_diff/list/' . $tablename) . '</em>';
        }
      }
    }
  }
}

/**
 * Implements hook_form_FORM_ID_alter().
 */
function _ctex_diff_form_feeds_ui_overview_form_alter(&$form, &$form_state) {

  foreach (array('enabled', 'disabled') as $state) {
    if (isset($form[$state])) {
      foreach ($form[$state] as $id => &$importer_form) {
        if ($importer_form['status']['#markup'] === t('Overridden')) {
          $importer_form['operations']['#markup'] .=
            ' | ' . l(t('Diff'), 'admin/structure/ctex_diff/diff/feeds_importer/' . $id)
            . ' | ' . l(t('View default'), 'admin/structure/ctex_diff/code/feeds_importer/' . $id);
        }
      }
    }
  }
}

/**
 * Implements hook_form_FORM_ID_alter().
 */
function _ctex_diff_form_path_breadcrumbs_ui_breadcrumbs_list_alter(&$form, &$form_state) {

  $result = path_breadcrumbs_load_all();
  foreach ($result as $path) {
    if ($path->export_type == (EXPORT_IN_CODE | EXPORT_IN_DATABASE)
        && isset($form[$path->machine_name]['actions']['#links'])) {
      $form[$path->machine_name]['actions']['#links'][] = array(
        'title' => t('Diff'),
        'href' => 'admin/structure/ctex_diff/diff/path_breadcrumbs/' . $path->machine_name,
      );
      $form[$path->machine_name]['actions']['#links'][] = array(
        'title' => t('View default'),
        'href' => 'admin/structure/ctex_diff/code/path_breadcrumbs/' . $path->machine_name,
      );
    }
  }
}

/**
 * Implements hook_form_FORM_ID_alter().
 */
function _ctex_diff_form_strongarm_admin_form_alter(&$form, &$form_state) {

  // $form has a specific format with its own theme function.
  // There are 3 different form vars containing the cells for 3 columns, as
  // #markup. The easiest is to cram a link into one of them.
  foreach ($form['storage'] as $name => &$storage) {
    if ($storage['#markup'] === t('Overridden')) {
      $storage['#markup'] .= ' ' . l(t('diff'), 'admin/structure/ctex_diff/diff/variable/' . $name);
    }
  }
}

/**
 * Implements hook_form_FORM_ID_alter().
 */
function _ctex_diff_form_entity_ui_overview_form_alter(&$form, &$form_state) {

  $entity_type = str_replace('_overview_form', '', $form['#form_id']);
  if ($entity_type == $form['#form_id']) {
    drupal_set_message(t('Form ID not of type X_overview_form? Not adding diff links.'), 'error');
    return;
  }
  $info = entity_get_info($entity_type);
  if (!$info) {
    drupal_set_message(t("Unknown entity type '@type' as derived from form ID. Not adding diff links.", array('@type' => $entity_type)), 'error');
    return;
  }

  if (!isset($form['table']['#rows']) || empty($form['table']['#header'])) {
    // This is the structure built by EntityDefaultUIController::overviewForm().
    drupal_set_message(t('table/#header/#rows structure not found in X_overview_form. Not adding diff links.'), 'error');
    return;
  }
  // Look for the header-column containing the 'Operations' links. It's multi-
  // colspan, at least in EntityDefaultUIController::overviewTableHeaders().
  // Assume the header definitions have zero-based numeric indexes.
  $new_data_col = 0;
  foreach ($form['table']['#header'] as $header_col => $header_def) {
    if (isset($header_def['data'])
        && $header_def['data'] === t('Operations')
        && !empty($header_def['colspan'])) {
      $new_data_col = $header_col + $header_def['colspan'];
      break;
    }
  }
  if (!$new_data_col) {
    // This is the structure built by EntityDefaultUIController::overviewForm().
    drupal_set_message(t('Operations header not found in X_overview_form. Not adding diff links.'), 'error');
    return;
  }

  // Get all 'config entities' and index them by machine name. (We probably
  // have them in the same order as $variables['rows'] but can't count on it.)
  $key = isset($info['entity keys']['name']) ? $info['entity keys']['name'] : $info['entity keys']['id'];
  $entities = entity_load($entity_type);
  $configs = array();
  foreach ($entities as $entity) {
    $configs[$entity->$key] = $entity;
  }

  $statuskey = isset($info['entity keys']['status']) ? $info['entity keys']['status'] : 'status';
  $overridden_rows = FALSE;
  $not_overridden_rows = array();
  // Check all rows, for the last column containing an 'edit' link and extract
  // the machine name.
  foreach ($form['table']['#rows'] as $row_key => $row) {

    // Machine name is stuck in data/#name by overviewTableRow().
    if (isset($row[0]['data']['#name']) && isset($configs[$row[0]['data']['#name']]->$statuskey)
        && $configs[$row[0]['data']['#name']]->$statuskey == ENTITY_OVERRIDDEN) {

      array_splice($form['table']['#rows'][$row_key], $new_data_col, 0, array(
        l(t('diff'), "admin/structure/ctex_diff/diff/$entity_type/" . $row[0]['data']['#name'], array('alias' => TRUE)),
        l(t('default'), "admin/structure/ctex_diff/code/$entity_type/" . $row[0]['data']['#name'], array('alias' => TRUE)),
      ));
      $overridden_rows = TRUE;
    }
    else {
      $not_overridden_rows[] = $row_key;
    }
  }
  if ($overridden_rows) {
    // Add 2 extra columns to each non-overridden row.
    $form['table']['#header'][$header_col]['colspan'] += 2;
    foreach ($not_overridden_rows as $row_key) {
      array_splice($form['table']['#rows'][$row_key], $new_data_col, 0, array('', ''));
    }
  }
}

/**
 * Implements hook_form_FORM_ID_alter().
 */
function _ctex_diff_form_facetapi_realm_settings_form_alter(&$form, &$form_state) {
  if (!isset($form['#facetapi']['realm']['name']) || !isset($form['#facetapi']['adapter'])) {
    watchdog('ctex_diff', 'Realm name not found in @function; code should be adjusted.', array('@function' => __FUNCTION__), WATCHDOG_ERROR);
    return;
  }
  $realm = $form['#facetapi']['realm']['name'];
  $adapter = $form['#facetapi']['adapter'];
  $searcher = $adapter->getSearcher();
  ctools_include('export');
  $configs = ctools_export_crud_load_all('facetapi');
  foreach ($form['#facetapi']['facet_info'] as $facet_name => $facet) {
    $key = "$searcher:$realm:$facet_name";
    if (isset($configs[$key]->export_type)
        && $configs[$key]->export_type == (EXPORT_IN_CODE | EXPORT_IN_DATABASE)) {
      $form['table']['operations'][$facet_name]['#links'][] = array(
        'title' => t('view default'),
        'href' => "admin/structure/ctex_diff/code/facetapi/$key",
      );
      $form['table']['operations'][$facet_name]['#links'][] = array(
        'title' => t('diff'),
        'href' => "admin/structure/ctex_diff/diff/facetapi/$key",
      );
    }
    // I'm not 100% sure about the following but the fact that the facetapi
    // module's "export" link export 2 separate objects, means that something
    // like this could be useful too...
    $key = "$searcher::$facet_name";
    if (isset($configs[$key]->export_type)
        && $configs[$key]->export_type == (EXPORT_IN_CODE | EXPORT_IN_DATABASE)) {
      $form['table']['operations'][$facet_name]['#links'][] = array(
        'title' => t('view default (general settings)'),
        'href' => "admin/structure/ctex_diff/code/facetapi/$key",
      );
      $form['table']['operations'][$facet_name]['#links'][] = array(
        'title' => t('diff (general settings)'),
        'href' => "admin/structure/ctex_diff/diff/facetapi/$key",
      );
    }
  }
}

/**
 * Implements hook_form_FORM_ID_alter().
 */
function _ctex_diff_form_search_api_autocomplete_admin_overview_alter(&$form, &$form_state) {
  foreach ($form['search_api_views']['searches']['#options'] as $key => &$option) {
    // This is a rendered item_list by default - meaning the only way to
    // insert extra links is in the HTML, making assumptions.
    if (preg_match('|<li class="last"|', $option['operations'], $m)) {
      $option['operations'] = str_replace(
        '<li class="last"',
        '<li>' . l(t('diff'), "admin/structure/ctex_diff/diff/search_api_autocomplete_search/$key", array('alias' => TRUE))
        . '</li><li>' . l(t('view default'), "admin/structure/ctex_diff/code/search_api_autocomplete_search/$key", array('alias' => TRUE))
        . '</li><li><li class="last"',
        $option['operations']
      );
    }
    else {
      watchdog('ctex_diff', 'Expected HTML matches not found in @path; code should be adjusted.', array('@path' => $_GET['q']), WATCHDOG_ERROR);
    }
  }
}

/**
 * Implements hook_preprocess_HOOK().
 */
function _ctex_diff_preprocess_table(&$variables) {
  static $seen;

  switch (strtolower($_GET['q'])) {
    // backup_migrate 'crud types' are built upon a common base class that is
    // hard/impossible to override, that builds a table with already-rendered
    // links, and then renders the table to HTML, inside its own callback. So we
    // can only override the table here. (Or the HTML in hook_page_alter... no.)
    // The only thing we have to recognize this table as a b-m crud type list,
    // is the URL - plus our assumption that the last row should contain links.
    case 'admin/config/system/backup_migrate/schedule':
      $variables['rows'] = _ctex_diff_add_links_to_flat_table_rows($variables['rows'], 'backup_migrate_schedules', '', TRUE, ' &nbsp; ');
      break;

    case 'admin/config/system/backup_migrate/settings/destination':
      $variables['rows'] = _ctex_diff_add_links_to_flat_table_rows($variables['rows'], 'backup_migrate_destinations', '', TRUE, ' &nbsp; ');
      break;

    case 'admin/config/system/backup_migrate/settings/profile':
      $variables['rows'] = _ctex_diff_add_links_to_flat_table_rows($variables['rows'], 'backup_migrate_profiles', '', TRUE, ' &nbsp; ');
      break;

    case 'admin/config/system/backup_migrate/settings/source':
      $variables['rows'] = _ctex_diff_add_links_to_flat_table_rows($variables['rows'], 'backup_migrate_sources', '', TRUE, ' &nbsp; ');
      break;

    case 'admin/config/system/backup_migrate/settings':
      // This screen holds 4 separate tables. Since there is no good property to
      // recognize them by, we have to resort to ugly tricks that may get
      // outdated unexpectedly.
      if ($variables['header'][1] === t('Destinations')) {
        $variables['rows'] = _ctex_diff_add_links_to_flat_table_rows($variables['rows'], 'backup_migrate_schedules', 'admin/config/system/backup_migrate/schedule', TRUE, ' &nbsp; ');
      }
      elseif ($variables['header'][1] === t('Source')) {
        $variables['rows'] = _ctex_diff_add_links_to_flat_table_rows($variables['rows'], 'backup_migrate_profiles', 'admin/config/system/backup_migrate/settings/profile', TRUE, ' &nbsp; ');
      }
      elseif ($variables['header'][2] === t('Location')) {
        // Tables for sources & destinations are exactly the same. Use static
        // toggle.
        if (empty($seen)) {
          $variables['rows'] = _ctex_diff_add_links_to_flat_table_rows($variables['rows'], 'backup_migrate_sources', 'admin/config/system/backup_migrate/settings/source', TRUE, ' &nbsp; ');
          $seen = TRUE;
        }
        else {
          $variables['rows'] = _ctex_diff_add_links_to_flat_table_rows($variables['rows'], 'backup_migrate_destinations', 'admin/config/system/backup_migrate/settings/destination', TRUE, ' &nbsp; ');
        }
      }
      break;

    // Data has a simple page callback. For the rest: same as above.
    case 'admin/structure/data':
      $variables['rows'] = _ctex_diff_add_links_to_flat_table_rows($variables['rows'], 'data_tables');
      break;

    // Same for ds_view_modes/ds_fields.
    case 'admin/structure/ds/fields':
    case 'admin/structure/ds/view_modes':
      $type = strtolower(substr($_GET['q'], strlen('admin/structure/ds/')));
      if (in_array($type, array('fields', 'view_modes'), TRUE)) {

        ctools_include('export');
        $configs = ctools_export_crud_load_all('ds_' . $type);
        // Check all rows for the last column containing an 'edit' link and
        // extract the machine name.
        foreach ($variables['rows'] as $row_key => $row) {
          if (preg_match('|^<a href="[^"]*?/admin/structure/ds/' . $type . '/manage/(\w+)">|', end($row), $m)) {
            $column_key = key($row);

            if (isset($configs[$m[1]]->export_type)
                && $configs[$m[1]]->export_type == (EXPORT_IN_CODE | EXPORT_IN_DATABASE)) {
              $variables['rows'][$row_key][$column_key] .= ' - ' . l(t('Diff'), 'admin/structure/ctex_diff/diff/ds_' . $type . '/' . $m[1], array('alias' => TRUE))
                . ' - ' . l(t('View default'), 'admin/structure/ctex_diff/code/ds_' . $type . '/' . $m[1], array('alias' => TRUE));
            }
          }
        }
      }
      break;

    // Same for OAuth contexts.
    case 'admin/config/services/oauth/contexts':
      // Check all rows, for the "operations][data" field containing an 'edit'
      // link and extract the machine name.
      $m = array();
      foreach ($variables['rows'] as $row_key => $row) {
        if (preg_match('|<a href="[^"]*?/admin/config/services/oauth/(\w+)/edit">|', $row['data']['operations']['data'], $m)) {
          // This is a rendered item_list by default - meaning the only way to
          // insert extra links is in the HTML, making assumptions. Let's not
          // act too difficult for now: assume theme_links() output a 'li'
          // element with its common classes and there are less than 10 links,
          // but do not add classes to the added HTML.
          $key = $m[1];
          if (preg_match('|<li class="(\d) last"|', $row['data']['operations']['data'], $m)) {
            $variables['rows'][$row_key]['data']['operations']['data'] = str_replace(
              "<li class=\"$m[1] last\"",
              '<li>' . l(t('Diff'), "admin/structure/ctex_diff/diff/oauth_common_context/$key", array('alias' => TRUE))
              . '</li><li>' . l(t('View default'), "admin/structure/ctex_diff/code/oauth_common_context/$key", array('alias' => TRUE))
              . "</li><li><li class=\"$m[1] last\"",
              $variables['rows'][$row_key]['data']['operations']['data']
            );
          }
          else {
            watchdog('ctex_diff', 'Expected HTML matches not found in @path; code should be adjusted.', array('@path' => $_GET['q']), WATCHDOG_ERROR);
          }
        }
      }
      break;

    // Metatag also builds its full table in a simple page callback, but returns
    // a render array instead of a HTML string. So we can override things here
    // or in hook_page_alter(), it doesn't matter much...
    // The Meta tags page renders several nested tables; the check for the
    // header makes sure we alter the right one.
    case 'admin/config/search/metatags':

      if (isset($variables['header']['operations'])
          && $variables['header']['operations'] === t('Operations')) {

        ctools_include('export');
        $configs = ctools_export_crud_load_all('metatag_config');
        foreach ($variables['rows'] as $row_key => $row) {
          if (isset($configs[$row_key]->export_type)
              && $configs[$row_key]->export_type == (EXPORT_IN_CODE | EXPORT_IN_DATABASE)
              // We assume there's at least one link present. Skip if not.
              && isset($row['data']['operations']['data']['#links'])) {

            $variables['rows'][$row_key]['data']['operations']['data']['#links']['diff'] =
              array('title' => t('Diff'), 'href' => 'admin/structure/ctex_diff/diff/metatag_config/' . $row_key);
            $variables['rows'][$row_key]['data']['operations']['data']['#links']['code'] =
              array('title' => t('View default'), 'href' => 'admin/structure/ctex_diff/code/metatag_config/' . $row_key);
          }
        }
      }
      break;

    case 'admin/config/search/search_api':
      $server_configs = entity_load_multiple_by_name('search_api_server');
      $index_configs = entity_load_multiple_by_name('search_api_index');
      $m = array();
      foreach ($variables['rows'] as $row_key => $row) {
        if (preg_match('|<a href="[^"]*?/admin/config/search/search_api/server/(\w+)/edit">|', end($row), $m)) {
          $column_key = key($row);
          if (isset($server_configs[$m[1]]->status)
              && $server_configs[$m[1]]->status == (EXPORT_IN_CODE | EXPORT_IN_DATABASE)) {
            // This is a rendered item_list by default - meaning the only way to
            // insert extra links is in the HTML, making assumptions. Let's not
            // act too difficult for now: assume theme_links() output a 'li'
            // element with its common classes, but do not add classes to the
            // added HTML.
            $key = $m[1];
            if (preg_match('|<li class="search-api-server-(\w+) last"|', $row[$column_key], $m)) {
              $variables['rows'][$row_key][$column_key] = str_replace(
                "<li class=\"search-api-server-$m[1] last\"",
                '<li>' . l(t('Diff'), "admin/structure/ctex_diff/diff/search_api_server/$key", array('alias' => TRUE))
                . '</li><li>' . l(t('View default'), "admin/structure/ctex_diff/code/search_api_server/$key", array('alias' => TRUE))
                . "</li><li><li class=\"search-api-server-$m[1] last\"",
                $variables['rows'][$row_key][$column_key]
              );
            }
            else {
              watchdog('ctex_diff', 'Expected HTML matches not found in @path; code should be adjusted.', array('@path' => $_GET['q']), WATCHDOG_ERROR);
            }
          }
        }
        elseif (preg_match('|<a href="[^"]*?/admin/config/search/search_api/index/(\w+)/edit">|', end($row), $m)) {
          $column_key = key($row);
          if (isset($index_configs[$m[1]]->status)
              && $index_configs[$m[1]]->status == (EXPORT_IN_CODE | EXPORT_IN_DATABASE)) {
            // This is a rendered item_list by default - meaning the only way to
            // insert extra links is in the HTML, making assumptions. Let's not
            // act too difficult for now: assume theme_links() output a 'li'
            // element with its common classes, but do not add classes to the
            // added HTML.
            $key = $m[1];
            if (preg_match('|<li class="search-api-index-(\w+) last"|', $row[$column_key], $m)) {
              $variables['rows'][$row_key][$column_key] = str_replace(
                "<li class=\"search-api-index-$m[1] last\"",
                '<li>' . l(t('Diff'), "admin/structure/ctex_diff/diff/search_api_index/$key", array('alias' => TRUE))
                . '</li><li>' . l(t('View default'), "admin/structure/ctex_diff/code/search_api_index/$key", array('alias' => TRUE))
                . "</li><li><li class=\"search-api-index-$m[1] last\"",
                $variables['rows'][$row_key][$column_key]
              );
            }
            else {
              watchdog('ctex_diff', 'Expected HTML matches not found in @path; code should be adjusted.', array('@path' => $_GET['q']), WATCHDOG_ERROR);
            }
          }
        }
      }
  }
}

/**
 * Adds diff/original links to cells in a table that adhere to some specifics.
 *
 * @param array $rows
 *   The table rows to add links to.
 * @param string $name
 *   The CTools object name.
 * @param string $path
 *   (optionaal) The (sub)path of an existing edit link to search for. If left
 *   empty, the current page's path will be taken.
 * @param bool $text_lowercase
 *   (optional) if TRUE, the text for the inserted links should be lowercase. If
 *   FALSE, they will have an uppercase first letter.
 * @param string $link_separator
 *   (optional) The separator between links used in the table cell.
 *
 * @return array
 *   The table rows with added links.
 */
function _ctex_diff_add_links_to_flat_table_rows(array $rows, $name, $path = '', $text_lowercase = FALSE, $link_separator = ' | ') {
  $text_diff = 'Diff';
  $text_code = 'View default';
  // We first lowercase and then translate the result, not the other way around.
  // Not because that's logical but because other links in the list will also be
  // doing it that way.
  if ($text_lowercase) {
    $text_diff = strtolower($text_diff);
    $text_code = strtolower($text_code);
  }
  if (!$path) {
    $path = strtolower($_GET['q']);
  }
  ctools_include('export');
  $configs = ctools_export_crud_load_all($name);

  // Check all rows for the last column containing an 'edit' link and
  // extract the machine name.
  $m = array();
  foreach ($rows as $row_key => $row) {
    // (One known case where the row does not _start_ with the edit link, so no
    // '^': backup_migrate_destinations.)
    if (preg_match('|<a href="[^"]*?/' . $path . '/edit/(\w+)">|', end($row), $m)) {
      $column_key = key($row);

      if (isset($configs[$m[1]]->export_type)
          && $configs[$m[1]]->export_type == (EXPORT_IN_CODE | EXPORT_IN_DATABASE)) {
        $rows[$row_key][$column_key] .= $link_separator . l(t($text_diff), "admin/structure/ctex_diff/diff/$name/$m[1]", array('alias' => TRUE))
          . $link_separator . l(t($text_code), "admin/structure/ctex_diff/code/$name/$m[1]", array('alias' => TRUE));
      }
    }
  }

  return $rows;
}

/**
 * Implements hook_page_alter().
 */
function _ctex_diff_page_alter(&$page) {

  // Rules- payment methods:
  // don't have a #action. The payment methods admin page (in newer commerce
  // versions) gave us a #page_callback value to hook into:
  if (isset($page['content']['system_main']['#page_callback'])
          && $page['content']['system_main']['#page_callback'] === 'commerce_payment_ui_admin_page') {

    // We have 2 tables:
    foreach (array('enabled', 'disabled') as $key) {
      _ctex_diff_alter_rules_table($page['content']['system_main'][$key]['rules'], $key);
    }
  }

  if (isset($page['content']['system_main']['#action'])) {

    // Other Rules screens:
    if ($page['content']['system_main']['#action'] === '/admin/config/workflow/rules/components') {

      _ctex_diff_alter_rules_table($page['content']['system_main']['table']);
    }
    elseif ($page['content']['system_main']['#action'] === '/admin/config/workflow/rules'
            || strpos($page['content']['system_main']['#action'], '/admin/config/workflow/rules?') === 0) {

      // We have 2 tables:
      foreach (array('active', 'inactive') as $key) {
        _ctex_diff_alter_rules_table($page['content']['system_main'][$key], $key);
      }
    }
  }
}

/**
 * Inserts extra links into a rules-admin like table.
 *
 * @param array $table_element
 *   Part of a renderable form element that is a table (having #rows)
 * @param string $id_string
 *   String only used for error reporting if necessary.
 */
function _ctex_diff_alter_rules_table(array &$table_element, $id_string = '') {
  // For the constants:
  ctools_include('export');

  $cols_added = FALSE;

  foreach ($table_element['#rows'] as $row_key => &$row) {

    // System name is buried inside a link. Cells 3-8 are links.
    if (empty($row[0]['data']['description']['settings']['machine_name']['#markup'])) {
      drupal_set_message(t('ctex_diff: unexpected data structure for payment method @id; not adding links.', array('@id' => $id_string . '/' . $row_key)), 'error');
    }
    else {
      $m = array();
      if (!preg_match('|: (\w+)|', $row[0]['data']['description']['settings']['machine_name']['#markup'], $m)) {
        drupal_set_message(t('ctex_diff: unable to extract Rule ID for payment method @id; not adding links.', array('@id' => $id_string . '/' . $row_key)), 'error');
      }
      elseif (!isset($row[2]['data']['#status'])) {
        drupal_set_message(t('ctex_diff: unexpected status data structure for payment method @id; not adding links.', array('@id' => $id_string . '/' . $row_key)), 'error');
      }
      else {
        $name = $m[1];
        $status = $row[2]['data']['#status'];
        if ($status == (EXPORT_IN_CODE | EXPORT_IN_DATABASE)) {
          $row[] = l(t('diff'), 'admin/structure/ctex_diff/diff/rules_config/' . $name);
          $row[] = l(t('default'), 'admin/structure/ctex_diff/code/rules_config/' . $name);
          if (!$cols_added && !empty($table_element['#header'][4]['colspan'])) {
            $table_element['#header'][4]['colspan'] += 2;
            $cols_added = TRUE;
          }
        }
        else {
          $row[] = '&nbsp;';
          $row[] = '&nbsp;';
        }
      }
    }
  }
}
